/*

  DBOX Module
  Copyright (C) 1994-2000 Daniel Kroening <kroening@handshake.de>

  Purpose: 

*/

#include <string>

#include <dblib/os_dep_process.h>
#include <tools.h>
#include <folder.h>
#include <stringtools.h>
#include <log_serv.h>
#include <logfile.h>
#include <dboxversion.h>
#include <messageid.h>

#ifndef DBOX_STANDALONE
#include <md5.h>
#endif

#include "pop.h"
#include "dboxpop3.h"

void pop3alarm(int signal)
 {
  std::cout << "-ERR logout, timeout\r\n";
  exit(0);
 }

class dboxpop3t:public dboxpopt
 {
 public:
  std::string command, args;
  std::vector<dword> messages;
  std::vector<bool> deleted;
  dword lastmsg;
  std::string timestamp;

  void dolist(dword arg, bool only_parameter, bool uidl);
  bool logged_in(); 
  void dopop3();
  void parse();
  void retr_top(dword nr, const char *toparg);

  dboxpop3t();
 };

dboxpop3t::dboxpop3t():dboxpopt()
 {
  protocol="pop3";
  lastmsg=0;
 }

void dboxpop3t::parse()
 {
  const char *tptr=commandline.c_str();

  while(*tptr==' ' || *tptr=='\t') tptr++;

  command="";
  while(*tptr!=0 && *tptr!=' ' && *tptr!='\t')
    command+=*(tptr++);

  if(*tptr==' ' || *tptr=='\t')
   {
    while(*tptr==' ' || *tptr=='\t') tptr++;
    args=tptr;
   }
  else
    args="";
 }

void dboxpop3t::retr_top(dword nr, const char *toparg)
 {
  dword lines;

  //
  // get header and, if necessary, body
  //

  if(toparg==NULL)
    lines=0xffffffff;
  else
    lines=atol(toparg);

  if(toparg==NULL)
    std::cout << "+OK message follows\r\n";
  else
    std::cout << "+OK top of message follows\r\n";

  if(lines!=0)
    get_body(nr);
  else
    get_header(nr);

  // output header

  std::cout.write(rfcm.data, rfcm.headerlen);

  // output body

  if(lines!=0)
   {
    const char *ptr=rfcm.data+rfcm.headerlen;
    char data=0, lastchar=0;
    byte escapestatus=1; // 0=nix, 1=Zeilenanfang, 2=Zeilenanfang+'.'

    while(lines!=0 && ptr<rfcm.endofmail)
     {
      lastchar=data;

      data=*(ptr++);

      switch(escapestatus)
       {
       case 0:
        if(data=='\r' || data=='\n')
          escapestatus=1;
        break;

       case 1:
        if(data=='.')
         {
          escapestatus=2;
          std::cout << '.'; // Hm. Netscape macht aus ... => ..
         }
        else if(data=='\r' || data=='\n')
          escapestatus=1;
        else
          escapestatus=0;
        break;

       case 2:
        if(data=='\r' || data=='\n')
          escapestatus=1;
        else
          escapestatus=0;
        break;
       }

      if(data=='\n' && lastchar!='\r')
        std::cout << '\r';

      std::cout << data;

      if(data=='\n') lines--;
     }
  
    if(data!='\n')
      std::cout << "\r\n";
   }  

  std::cout << ".\r\n";
 }

void dboxpop3t::dolist(dword arg, bool only_parameter, bool uidl)
 {
  if(only_parameter)
   {
    if(deleted[arg-1])
     {
      std::cout << "-ERR message is deleted\r\n";
      return;
     }

    std::cout << "+OK ";
   }
  else
   {
    if(deleted[arg-1])
      return;
   }

  dword nr=messages[arg-1];
  get_folderindex(nr);

  std::cout << arg << " ";

  if(uidl)
   {
    std::cout << fi.id << "."
              << str2crc(fi.mid) << "."
              << fi.received;
   }
  else
   {
    if(!fi.flags.binary)
      std::cout << (fi.len+fi.headerlen);
    else
      std::cout << (fi.headerlen+fi.len*1377/1000);
   }

  std::cout << "\r\n";
 }

bool dboxpop3t::logged_in()
 {
  if(open_mail(up.name, ""))
   {
    std::cout << "-ERR internal error: failed to open mail folder\r\n";
    return true;
   }

  if(index>=0 && header>=0)
   {
    dword idx=0;

    while(read(index, &fi, sizeof fi)==sizeof fi)
     {
      if(fi.check_version())
       {
        log_serv.printf("dboxpopt::login", "mailbox %s corrupted",
          up.name);

        std::cout << "-ERR sorry, mailbox is corrupted\r\n";
        abort();
       }

      if(!fi.flags.deleted)
        messages.push_back(idx);

      idx++;
     }

    deleted.resize(messages.size(), false);
   }

  return false;
 }

#ifndef DBOX_STANDALONE
std::string correct_digest(const std::string &timestamp, const std::string &password)
 {
  std::string hash_input=timestamp+password;

  unsigned char md5hash[16];
  md5_calc(md5hash, (unsigned char *)hash_input.c_str(), hash_input.size());

  std::string digest;

  for(int i=0; i<16; i++)
   {
    char hex_s[4];
    snprintf(hex_s, 3, "%02x", (unsigned char)md5hash[i]);
    digest+=hex_s;
   }

  return digest;
 }
#endif

void dboxpop3t::dopop3()
 {
  signal(SIGALRM, pop3alarm);

  timestamp="<"+gen_message_id("pop3")+">";
  std::cout << "+OK DBOX POP3 Server " DBOXVERSIONNR " ready";

  #ifndef DBOX_STANDALONE
  std::cout << " " << timestamp;
  #endif

  std::cout << "\r\n";

  getpeer();

  while(true)
   {
    if(getcommand()) return;
    parse();

    if(stringcasecmp(command, "QUIT")==0)
     {
      if(fp.lock())
        std::cout << "-ERR sorry, failed to get a lock\r\n";
      else
       {
        bool crunch=false;

        for(unsigned i=0; i<messages.size(); i++)
          if(deleted[i])
           {
            dword nr=messages[i];
            lseek(index, nr*sizeof(fi), SEEK_SET);
            read(index, &fi, sizeof fi);

            if(!fi.flags.deleted)
             {
              fi.flags.deleted=true;
              lseek(index, nr*sizeof(fi), SEEK_SET);
              write(index, &fi, sizeof fi);
              crunch=true;
             }
           }

        fp.unlock();

        std::cout << "+OK " << config.boxname
                  << " DBOX POP Server (Version vom " DBOXVERSION ") shutdown.\r\n";

        #ifndef DBOX_STANDALONE // TODO
        if(crunch)
          docrunch(up.name);
        #endif

        return;
       }
     }
    else if(stringcasecmp(command, "USER")==0 && !transfer)
     {
      up.setname(args);
      if(up.name[0]==0)
        std::cout << "-ERR invalid user name\r\n";
      else
       {
        dnstring(up.name);
        std::cout << "+OK please send PASS command\r\n";
        #ifndef DBOX_STANDALONE
        snprintf(userlogfile, DIRLEN, "%slogfile", up.getpath());
        #endif
       }
     }
    else if(stringcasecmp(command, "PASS")==0 && !transfer)
     {
      if(up.name[0]==0)
       {
        std::cout << "-ERR please provide an user name first\r\n";
       }
      else
       {
        if(login(args))
         {
          std::cout << "-ERR invalid password, bye\r\n";
          return;
         }

        if(logged_in())
          return;

        std::cout << "+OK Welcome\r\n";
       }
     }
    else if(stringcasecmp(command, "STAT")==0 && transfer)
     {
      if(index<0)
        std::cout << "+OK 0 0\r\n";
      else
       {
        dword total_size=0; // estimated

        for(dword arg=1; arg<=messages.size(); arg++)
         {
          if(!deleted[arg-1])
           {
            get_folderindex(messages[arg-1]);

            if(!fi.flags.binary)
              total_size+=fi.len+fi.headerlen;
            else
              total_size+=fi.headerlen+fi.len*1377/1000; // uuencode
           }
         }

        std::cout << "+OK " << messages.size() << " " 
                            << total_size << "\r\n";
       }
     }
    else if(stringcasecmp(command, "LIST")==0 && transfer)
     {
      if(args=="")
       {
        std::cout << "+OK " << messages.size() << " messages\r\n";

        for(dword arg=1; arg<=messages.size(); arg++)
          dolist(arg, false, false);

        std::cout << ".\r\n";
       }
      else
       {
        dword arg=atoi(args.c_str());
        if(arg==0 || arg>messages.size())
          std::cout << "-ERR invalid message\r\n";
        else
          dolist(arg, true, false);
       }
     }
    else if(stringcasecmp(command, "UIDL")==0 && transfer)
     {
      if(args=="")
       {
        std::cout << "+OK " << messages.size() << " messages\r\n";

        for(dword arg=1; arg<=messages.size(); arg++)
          dolist(arg, false, true);

        std::cout << ".\r\n";
       }
      else
       {
        dword arg=atoi(args.c_str());
        if(arg==0 || arg>messages.size())
          std::cout << "-ERR invalid message\r\n";
        else
          dolist(arg, true, true);
       }
     }
    else if((stringcasecmp(command, "RETR")==0 ||
            stringcasecmp(command, "TOP")==0) && transfer)
     {
      const char *toparg=(char *)NULL;

      if(stringcasecmp(command, "TOP")==0)
       {
        toparg=strrchr(args.c_str(), ' ');
        if(toparg!=NULL) toparg++;
       }

      if(args=="")
        std::cout << "-ERR message number required!\r\n";
      else
       {
        dword arg, nr;

        arg=atoi(args.c_str());
        if(arg==0 || arg>messages.size())
          std::cout << "-ERR invalid message\r\n";
        else if(deleted[arg-1])
          std::cout << "-ERR message is deleted\r\n";
        else
         {
          if(arg>lastmsg) lastmsg=arg;
          nr=messages[arg-1];
          retr_top(nr, toparg);
         }
       }
     }
    else if(stringcasecmp(command, "DELE")==0 && transfer)
     {
      if(args=="")
        std::cout << "-ERR message number required!\r\n";
      else
       {
        dword arg=atoi(args.c_str());

        if(arg==0 || arg>messages.size())
          std::cout << "-ERR invalid message\r\n";
        else if(deleted[arg-1])
          std::cout << "-ERR message is deleted\r\n";
        else
         {
          if(arg>lastmsg) lastmsg=arg;
          deleted[arg-1]=true;
          std::cout << "+OK message deleted.\r\n";
         }
       }
     }
    else if(stringcasecmp(command, "NOOP")==0 && transfer)
     {
      if(args!="")
        std::cout << "-ERR Uh?\r\n";
      else
        std::cout << "+OK Gaehn.\r\n";
     }
    else if(stringcasecmp(command, "APOP")==0)
     {
      #ifdef DBOX_STANDALONE
      std::cout << "-ERR sorry, no APOP supported\r\n";
      #else
      char mailbox[TEXTLEN], *tptr;
      strmaxcpy(mailbox, args.c_str(), TEXTLEN-1);
      tptr=strchr(mailbox, ' ');
      if(tptr==NULL)
        std::cout << "-ERR invalid paramters\r\n";
      else
       {
        *tptr=0;
        const char *digest=strchr(args.c_str(), ' ');
        if(digest!=NULL) digest++; else digest="";

        dnstring(mailbox);
        up.setname(mailbox);

        if(up.read())
         {
          log_serv.printf("login", "Invalid user name %s from %s", up.name, peer.c_str());
          std::cout << "-ERR invalid user name\r\n";
         }
        else
         {
          snprintf(userlogfile, DIRLEN, "%slogfile", up.getpath());

          const char *cryptedpassword=up|"cryptedpassword";
          if(cryptedpassword[0]==0) cryptedpassword=up|"cryptedpasswort";
          if(cryptedpassword[0]==0 || strcmp(cryptedpassword, "locked")==0)
           {
            log_serv.printf("login", "apop: account %s locked from %s", up.name, peer.c_str());
            std::cout << "-ERR account locked\r\n";
           }
          else
           {
            char password[TEXTLEN];
            uncryptpassword(cryptedpassword, password);

            if(correct_digest(timestamp, password)==digest)
             {
              log_serv.printf("login", "%s: correct APOP digest from %s", up.name, peer.c_str());

              if(login2())
                std::cout << "-ERR permission denied\r\n";
              else if(logged_in())
                return;
              else
                std::cout << "+OK Welcome\r\n";
             }
            else
             {
              logfileprintf(userlogfile, "login", 
                "wrong %s APOP digest from %s", protocol.c_str(), peer.c_str());
              log_serv.printf("login", "%s: wrong APOP digest from %s", up.name, peer.c_str());
              std::cout << "-ERR invalid password\r\n";
             }
           }
         }
       }
      #endif
     }
    else if(stringcasecmp(command, "LAST")==0 && transfer)
     {
      if(args!="")
        std::cout << "-ERR Uh?\r\n";
      else
        std::cout << "+OK " << lastmsg << "\r\n";
     }
    else if(stringcasecmp(command, "RSET")==0 && transfer)
     {
      if(args!="")
        std::cout << "-ERR Uh?\r\n";
      else
       {
        for(unsigned i=0; i<deleted.size(); i++)
          deleted[i]=false;

        std::cout << "+OK\r\n";
       }
     }
    else
      std::cout << "-ERR Invalid command: \"" << command << "\"\r\n";

    std::cout.flush();
   }
 }

void dopop3()
 {
  dboxpop3t dboxpop3;

  dboxpop3.dopop3();

  #if 0
  // APOP Example from rfc1939
  std::cout << 
   correct_digest("<1896.697170952@dbc.mtview.ca.us>", "tanstaaf") 
   << "\n";
  #endif
 }
