/*

  DBOX Module
  Copyright (C) 1994-2000 Daniel Kroening <kroening@handshake.de>

  Purpose: 

*/

#include <fstream>

#include <libdbox/general.h>
#include <libdbox/stringtools.h>
#include <libdbox/textprofile.h>
#include <dblib/rfc1522.h>
#include <dblib/file_table.h>
#include <dblib/os_dep_file.h>
#include <dbclient/logstr.h>
#include <dbclient/client.h>

#include "deliver.h"
#include "spool_in_folder.h"
#include "folder_path.h"

/*******************************************************************\

Function: match_header

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

bool match_header(const rfcmessaget &m,
                  const std::string &header,
                  const std::string &pattern)
 {
  const char *start=m.data, *text;
  bool mime_decode=false;

  if(stringcasecmp(header, "From")==0 ||
     stringcasecmp(header, "Subject")==0)
    mime_decode=true;

  while(!m.findheader(&start, header.c_str(), &text))
   {
    std::string header_text;
    m.copyheadertext(header_text, text);

    if(mime_decode)
      header_text=decode_rfc1522(header_text);

    if(check_joker(pattern, header_text, true))
      return true;
   }

  return false;
 }

/*******************************************************************\

   Class: deliver_mailt

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

class deliver_mailt
 {
 public:
  rfcmessaget &m;
  const std::string &user;
  const std::string &envelope_to, &envelope_from;
  std::string dest_user;
  std::string userlogfile;
  bool spam;

  bool has_maps;

  void deliver_mail();
  bool match_mail(textprofilet &profile);
  bool doit(textprofilet &profile);
  void doalias(const std::string &alias);

  deliver_mailt(rfcmessaget &_m, 
                const std::string &_user,
                const std::string &_envelope_to,
                const std::string &_envelope_from,
                bool _spam);

 protected:
  std::string maildir;
 };

/*******************************************************************\

Function: deliver_mailt::deliver_mailt

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

deliver_mailt::deliver_mailt(
  rfcmessaget &_m,
  const std::string &_user,
  const std::string &_envelope_to,
  const std::string &_envelope_from,
  bool _spam):
  m(_m), user(_user),
  envelope_to(_envelope_to),
  envelope_from(_envelope_from),
  spam(_spam)
 {
 }

/*******************************************************************\

Function: deliver_mailt::match_mail

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

bool deliver_mailt::match_mail(textprofilet &profile)
 {
  for(unsigned i=0; i<profile.size(); i++)
   {
    if(strnicmp(profile.data[i].id, "header.", 7)==0)
     {
      if(!match_header(m, profile.data[i].id+7, profile.data[i].data))
        return false;
     }
    else if(stringcasecmp(profile.data[i].id, "envelope.to")==0)
     {
      if(!checkjoker(profile.data[i].data, envelope_to.c_str(), true))
        return false;
     }
    else if(stringcasecmp(profile.data[i].id, "envelope.from")==0)
     {
      if(!checkjoker(profile.data[i].data, envelope_from.c_str(), true))
        return false;
     }
    else if(stringcasecmp(profile.data[i].id, "spam")==0)
     {
      if(profile.data[i].data!=i2string(spam))
        return false;
     }
   }
  
  return true; 
 }

/*******************************************************************\

Function: deliver_mailt::doalias

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

void deliver_mailt::doalias(const std::string &alias)
 {
  deliver_mailt deliver_mail(m, alias, envelope_to, envelope_from, spam);
  deliver_mail.deliver_mail(); // TODO: loops
 }

/*******************************************************************\

Function: deliver_mailt::doalias

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

bool deliver_mailt::doit(textprofilet &profile)
 {
  bool do_break=(atoi(profile.get("break"))!=0);
  std::string alias=profile.get("alias");

  if(alias!="")
   {
    doalias(alias);
    return do_break;
   }

  std::string mailbox_name=profile.get("mailbox"),
              folder_owner=profile.get("owner"),
              subfolder;

  if(folder_owner=="")
    folder_owner=dest_user;  // default is destination user

  if(mailbox_name=="")
    return do_break;

  spool_in_folder(m, folder_owner, mailbox_name,
                  false, envelope_to, envelope_from);

  return do_break;
 }

/*******************************************************************\

Function: deliver_mailt::doalias

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

void deliver_mailt::deliver_mail()
 {
  maildir=client.config["maildir"];

  if(maildir=="")
    throw excpt("configuration error: maildir not set");

  std::string user_path=build_path(maildir, user);

  userlogfile=build_path(user_path, "logfile");

   {
    std::string filename=build_path(user_path, "emailtab");
    file_tablet email_table(filename, table_baset::MAP);
    dest_user=email_table[envelope_to];
    if(dest_user.empty()) dest_user=user;
   }

  user_path=build_path(maildir, dest_user);

  textprofilet profile;
  std::ifstream in((build_path(user_path, "mfilter")).c_str());

  bool do_default=true;

  if(in)
   {
    while(in && !in.eof())
      if(!profile.readline(in))
        if(match_mail(profile))
          if(doit(profile))
           {
            do_default=false;
            break;
           }
   }

  if(do_default)
   {
    if(spam && folder_exists(maildir, dest_user, SPAM_FOLDER))
      spool_in_folder(m, dest_user, SPAM_FOLDER,
                      false, envelope_to, envelope_from);
    else
      spool_in_folder(m, dest_user, "INBOX",
                      false, envelope_to, envelope_from);
   }
 }

/*******************************************************************\

Function: deliver

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

void deliver(rfcmessaget &m,
             const std::string &user,
             const std::string &envelope_to,
             const std::string &envelope_from,
             bool spam)
 {
  deliver_mailt deliver_mail(m, user, envelope_to, envelope_from, spam);

  deliver_mail.deliver_mail();
 }

