#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include <stdlib.h>
#include <time.h>
#include <sys/stat.h>
#include <signal.h>
#include <unistd.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <dirent.h>
#include <fcntl.h>
#include <arpa/inet.h>
#include <netdb.h>

#include <fstream>
#include <vector>

#include <log_serv.h>
#include <tools.h>
#include <dboxpath.h>
#include <lock.h>
#include <userdb.h>
#include <dbox.h>
#include <dblib/cmdline.h>
#include <dblib/os_dep_file.h>
#include <dblib/stream_copy.h>
#include <spoolp.h>
#include "netclient.h"

class netclientt
 {
 public:
  FILE *socket;
  int s;
  char response[TEXTLEN];
  userpt up;
  
  void openconnection(const char *hostname, const char *port);
  void donetclient(cmdlinet *cmdline);
  void login(const std::string &account);  
  void getresponse();
  void sendstring(const char *string);
  void sendspoolin(bool mail, const char **systems);
  void sendspoolin(bool mail, const char **systems, const char *directory);
  void sendspoolin(bool mail, const char **systems, const char *directory, const char *filename);
  void fetch(bool mail, bool maps, bool profiledata, bool erfc, const std::vector<std::string> &systems);
  void fetch(bool mail, bool maps, bool profiledata, bool erfc, const std::string &system);
  void fetchlist(bool mail, bool maps, bool profiledata, bool erfc);
 };
 
void netclientalarm(int signal)
 {
  printf("timeout, quit\r\n");
  exit(0);
 }

void netclientt::openconnection(const char *hostname, const char *port)
 {
  struct in_addr defaddr;
  struct hostent *hp, def;
  struct sockaddr_in sn;
  char *alist[1];

  printf("Opening connection\r\n");

  signal(SIGALRM, netclientalarm);
  signal(SIGPIPE, netclientalarm);
  alarm(5);

  if(hostname==NULL)
   {
    printf("No hostname, quit\r\n");
    exit(0);
   }

  if(port==NULL || port[0]==0)
    port="8081";

  hp=gethostbyname(hostname);

  if(!hp)
   {
    if(!inet_aton(hostname, &defaddr))
     {
      printf("unknown host: %s\r\n", hostname);
      exit(0);
     }

    def.h_name=(char *)hostname;
    def.h_addr_list=alist;
    def.h_addr=(char *)&defaddr;
    def.h_length=sizeof(struct in_addr);
    def.h_addrtype=AF_INET;
    def.h_aliases=0;
    hp=&def;
   }

  memset(&sn, 0, sizeof(sn));
  sn.sin_family = hp->h_addrtype;

  if(hp->h_length>(int)sizeof(sn.sin_addr))
    hp->h_length = sizeof(sn.sin_addr);

  memcpy(&sn.sin_addr, hp->h_addr, hp->h_length);
  sn.sin_port=htons(atoi(port));

  if((s=::socket(hp->h_addrtype, SOCK_STREAM, 0))<0)
    exit(0);

  if(connect(s, (struct sockaddr *)&sn, sizeof(sn))<0) 
   {
    printf("Connect failed.\r\n");
    log_serv.printf("openconnection", "Connect to %s:%s failed",
      hostname, port);
    exit(0);
   }

  socket=fdopen(s, "rw");

  if(!socket) exit(0);

  setvbuf(socket, NULL, _IONBF, 0);
 }

void netclientt::getresponse()
 {
  char *tptr;

  fflush(socket);

  alarm(30);
  if(fgets(response, TEXTLEN-1, socket)==NULL)
   {
    printf("timeout, quit\r\n");
    exit(0);
   }

  tptr=strchr(response, '\r'); if(tptr!=NULL) *tptr=0;
  tptr=strchr(response, '\n'); if(tptr!=NULL) *tptr=0;

  //printf("GOT: %s\r\n", response);
 }

void netclientt::sendstring(const char *string)
 {
  write(s, string, strlen(string));
 }

void netclientt::login(const std::string &account)
 {
  char password[TEXTLEN];

  printf("Login\r\n");

  strmaxcpy(up.name, account.c_str(), TEXTLEN-1);
  dnstring(up.name);

  if(up.read())
   {
    printf("   user %s not found\r\n", account.c_str());
    exit(0);
   }

  getresponse();

  if(response[0]!='+')
   {
    printf("   login failed\r\n"
           "   GOT: %s\r\n", response);
    log_serv.printf("login", "Login for %s failed", up.name);
    exit(0);
   }

  printf("   Sending username\r\n");
  sendstring("user ");
  sendstring(up.name);
  sendstring("\r\n");

  getresponse();

  if(response[0]!='+')
   {
    printf("   login failed\r\n"
           "   GOT: %s\r\n", response);
    exit(0);
   }

  getplaintextpassword(&up, password);
  printf("   Sending password\r\n");
  sendstring("pass ");
  sendstring(password);
  sendstring("\r\n");

  getresponse();

  if(response[0]!='+')
   {
    printf("   login failed\r\n"
           "   GOT: %s\r\n", response);
    exit(0);
   }

  printf("   Login OK\r\n");
 }

void netclientt::sendspoolin(bool mail, const char **systems)
 {
  char directory[DIRLEN];

  sprintf(directory, BOXSPOOLINZCMAILDIR, config.boxdir);
  sendspoolin(mail, systems, directory);
 }

void netclientt::sendspoolin(bool mail, const char **systems,
                             const char *directory,
                             const char *filename)
 {
  char username[TEXTLEN], *tptr;
  char outfilename[DIRLEN], tempstr[1000];
  dword size;
  int i;

  sprintf(outfilename, "%s%s", directory, filename);

  strmaxcpy(username, filename, TEXTLEN-1);
  tptr=strrchr(username, '.');
  if(tptr!=NULL) *tptr=0;

  for(i=0; systems[i]!=NULL; i++)
    if(stricmp(systems[i], username)==0 || 
       stricmp(systems[i], "*")==0)
      break;

  if(systems[i]==NULL) return;

  printf("\r\n   Sende Puffer von %s: %s\r\n", username, outfilename);

  size=filelength(outfilename);

  if(size>0)
   {
    std::ifstream infile(outfilename);

    if(!infile)
     {
      printf("Failed to open %s\r\n", outfilename);
      exit(0);
     }

    sprintf(tempstr, "transfer put zconnect %s source %s size %lu",
      mail?"mail":"news", username, size);

    printf("   Command: %s\r\n", tempstr);
    sendstring(tempstr);
    sendstring("\r\n");

    getresponse();
    if(response[0]!='+')
     {
      printf("   transfer command failed\r\n"
             "   GOT: %s\r\n", response);
      log_serv.printf("sendspoolin", "Transfer command failed: %s",
        response);
      exit(0);
     }

    printf("   Sending %lu bytes\r\n",  size);

    alarm(0);    
    filecopy(infile, s, size);
    
    getresponse();
    if(strncasecmp(response, "+OK rec", 6)!=0)
     {
      printf("   data transfer failed\r\n"
             "   GOT: %s\r\n", response);
      log_serv.printf("sendspoolin", "Failed to send %lu bytes for %s",
        size, username);
      exit(0);
     }

    log_serv.printf("sendspoolin", "Successfully sent %lu bytes for %s",
      size, username);
    printf("   Successfully sent %lu bytes\r\n", size);
   }

  unlink(outfilename);
 }

void netclientt::sendspoolin(bool mail, const char **systems, 
                             const char *directory)
 {
  char filename[DIRLEN];
  DIR *dir;
  dirent *ent;

  sprintf(filename, "%s.", directory);

  dir=opendir(filename);
  if(dir==NULL) return;

  while((ent=readdir(dir))!=NULL)
   if(ent->d_name[0]!='.')
     sendspoolin(mail, systems, directory, ent->d_name);

  closedir(dir);
 }

void netclientt::fetch(bool mail, bool maps, bool profiledata, bool erfc,
                       const std::string &system)
 {
  char tempstr[1000], outfilename[DIRLEN], username[TEXTLEN],
       newoutfilename[DIRLEN];
  dword size;
  int outfd;
  strmaxcpy(username, system.c_str(), TEXTLEN-1);
  dnstring(username);

  userpt up2(username);
  if(up2.read())
   {
    printf("     ... skipping %s, user not found\r\n", username);
    return;
   }

  if(profiledata)
   {
    if(stricmp(up2.get("gruppe"), "sysop")==0)
     {
      printf("     ... skipping %s, is sysop\r\n", username);
      return;
     }

    sprintf(tempstr, "transfer get profile source %s", username);

    printf("   Command: %s\r\n", tempstr);

    sendstring(tempstr);
    sendstring("\r\n");

    getresponse();
    if(strncmp(response, "+OK", 3)!=0)
     {
      printf("   transfer command failed\r\n"
             "   GOT: %s\r\n", response);
      close(outfd);
      unlink(outfilename);
      return;
     }

    std::vector<std::pair<std::string, std::string> > changes;

    while(TRUE)
     {
      getresponse();
      if(response[0]==0) break;
      char *tptr=strchr(response, '=');
      if(tptr!=NULL)
       {
        *(tptr++)=0;
        if(strcmp(up2.get(response), tptr)!=0)
         {
          printf("     setting %s=%s\r\n", response, tptr);
          changes.push_back(std::pair<std::string, std::string>(response, tptr));
         }
       }
     }

    if(!changes.empty())
     {
      if(up2.lock())
        printf("     failed to lock %s\r\n", up2.name);
      else
       {
        for(std::vector<std::pair<std::string, std::string> >::iterator it=changes.begin();
            it!=changes.end();
            it++)
          up2.set(it->first.c_str(), it->second.c_str());

         if(up2.write())
           printf("     failed to write profile of user %s\r\n", up2.name);
           
         up2.unlock();
       }
     }

    return;
   }

  printf("\r\n   Fetching new data from %s\r\n", username);

  sprintf(outfilename, DBOXTEMPDIR "netd.%d.%s", config.dboxdir, getpid(), username);
  outfd=open(outfilename, O_WRONLY|O_CREAT|O_TRUNC, 0600);
  if(outfd<0)
   {
    printf("   failed to open outfile\r\n");
    return;
   }

  sprintf(tempstr, "transfer get %s %s source %s", 
          erfc?"erfc":"zconnect",
          mail?"mail":(maps?"maps":"news"), username);

  printf("   Command: %s\r\n", tempstr);

  sendstring(tempstr);
  sendstring("\r\n");

  getresponse();
  if(strncmp(response, "+OK ", 4)!=0)
   {
    printf("   transfer command failed\r\n"
           "   GOT: %s\r\n", response);
    close(outfd);
    unlink(outfilename);
    return;
   }

  size=atol(response+4);

  printf("   Fetching %lu bytes\r\n", size);

  alarm(0); 

  if(filecopy(s, outfd, size))
   {
    printf("   data transfer failed\r\n");
    close(outfd);
    unlink(outfilename);
    return;
   }

  close(outfd);

  sendstring("+OK data received\r\n");
  printf("   Ok, data received\r\n");

  log_serv.printf("fetch", "Successfully fetched %lu bytes for %s", 
    size, username);

  if(maps)
   {
    sprintf(newoutfilename, BOXUSERDIR "maps", 
      config.boxdir, username);
    rename(outfilename, newoutfilename);
   }
  else
   {
    const char *lockname;

    if(erfc)
      lockname="puffer.erfc";
    else
      lockname=mail?"puffer.prv":"puffer.brt";

    while(lock(lockname, username, config.boxname, "fetch"))
     {
      printf("   locked\r\n");
      sleep(10);
     }

    if(erfc)
      sprintf(newoutfilename, BOXSPOOLDIR "%s.%s", 
        config.boxdir, username, username, "erfc");
    else
     sprintf(newoutfilename, BOXSPOOLDIR "%s.%s", 
        config.boxdir, username, username, mail?"prv":"brt");

    std::ifstream infile(outfilename);
    std::ofstream outfile(newoutfilename, std::ios::out|std::ios::binary|std::ios::app);

    if(!infile || !outfile)
     {
      printf("   internal error\r\n");
     }
    else
     {
      filecopy(infile, outfile, size);
      infile.close();
      outfile.close();
      unlink(outfilename);
     }

    unlock(lockname, username, config.boxname);
   }
 }

#define MAXFETCHLIST 10000

void netclientt::fetchlist(bool mail, bool maps, bool profiledata, bool erfc)
 {
  char *fetchlist[MAXFETCHLIST], tempstr[1000];
  int anz=0, i;

  printf("\r\n   Fetching system list\r\n");

  if(profiledata)
   {
    std::string userdir;
    getboxuserdir(userdir, config.boxdir, ".");
    userdir+=".";

    DIR *dir=opendir(userdir.c_str());

    if(dir==NULL)
     {
      fprintf(stderr, "failed to open \"%s\"\n", userdir.c_str());
      return;
     }

    struct dirent *ent;
    while((ent=readdir(dir))!=NULL)
      if(ent->d_name[0]!='.' && ent->d_name[0]!='_')
        if(anz<MAXFETCHLIST)
         {
          fetchlist[anz]=strdup(ent->d_name);
          if(fetchlist[anz]==NULL) abort();
          anz++;
         }

    closedir(dir);
   }
  else
   {
    sprintf(tempstr, "transfer get list %s %s", 
            erfc?"erfc":"zconnect",
            mail?"mail":(maps?"maps":"news"));

    printf("   Command: %s\r\n", tempstr);

    sendstring(tempstr);
    sendstring("\r\n");

    getresponse();
    if(strncmp(response, "+OK ", 4)!=0)
     {
      printf("   transfer command failed\r\n"
             "   GOT: %s\r\n", response);
      return;
     }

    while(TRUE)
     {
      getresponse();
      if(response[0]=='.') break;
      if(anz<MAXFETCHLIST)
       {
        fetchlist[anz]=strdup(response);
        if(fetchlist[anz]==NULL) abort();
        anz++;
       }
     }
   }

  printf("   Got %d system names\r\n", anz);

  for(i=0; i<anz; i++)
    fetch(mail, maps, profiledata, erfc, fetchlist[i]);

  for(i=0; i<anz; i++)
    free(fetchlist[i]);
 }

void netclientt::fetch(bool mail, bool maps, bool profiledata,
                       bool erfc, const std::vector<std::string> &systems)
 {
  for(unsigned i=0; i<systems.size(); i++)
    if(systems[i]=="*")
      fetchlist(mail, maps, profiledata, erfc);
    else
      fetch(mail, maps, profiledata, erfc, systems[i]);
 }

void netclientt::donetclient(cmdlinet *cmdline)
 {
  const char *account, *tptr;
  bool mail=FALSE, news=FALSE, maps=FALSE, profiledata=FALSE,
       erfc=FALSE;

  tptr=cmdline->getval('t');
  if(tptr==NULL || tptr[0]==0)
    return;
  else if(stricmp(tptr, "zcmail")==0)
    mail=TRUE;
  else if(stricmp(tptr, "erfcmail")==0)
   {
    mail=TRUE;
    erfc=TRUE;
   }
  else if(stricmp(tptr, "news")==0)
    news=TRUE;
  else if(stricmp(tptr, "maps")==0)
    maps=TRUE;
  else if(stricmp(tptr, "profile")==0)
    profiledata=TRUE;
  else
    return;

  account=cmdline->getval('s');
  if(account!=NULL && account[0]!=0)
   {
    openconnection(cmdline->getval('n'), cmdline->getval('p'));

    login(account);

    if(maps)
     {
      printf("\r\nFetching new data...\r\n");
      fetch(FALSE, TRUE, FALSE, erfc, cmdline->args);
     }
    else if(mail || news)
     {
      printf("\r\nSending spoolin...\r\n");
      sendspoolin(mail, (const char **)&cmdline->args);

      printf("\r\nFetching new data...\r\n");
      fetch(mail, FALSE, FALSE, erfc, cmdline->args);
     }
    else if(profiledata)
     {
      printf("\r\nUpdating profiles...\r\n");
      fetch(FALSE, FALSE, TRUE, FALSE, cmdline->args);
     }

    sendstring("QUIT\r\n");
   }
  else
   {
    if(cmdline->args.size()!=1)
      printf("Error: Specify only one system\r\n");

    printf("Zielsystem: %s\r\n", cmdline->args[0].c_str());

    openconnection(cmdline->getval('n'), cmdline->getval('p'));

    login(cmdline->args[0]);
   }

  fclose(socket);
 }

void donetclient(cmdlinet *cmdline)
 {
  netclientt netclient;

  netclient.donetclient(cmdline);
 }
